@@ -3,7 +3,6 @@
 require 'uri'
 require 'open-uri'
 require 'zlib'
-require 'open-uri'
 require 'net/http'
 
 module Asciidoctor
@@ -11,8 +10,8 @@ module Asciidoctor
   module PlantUml
     # PlantUML Configuration
     class Configuration
-      DEFAULT_URL = ENV['PLANTUML_URL'] || ''
-      DEFAULT_ENCODING = ENV['PLANTUML_ENCODING'] || 'legacy'
+      DEFAULT_URL = ENV.fetch('PLANTUML_URL', '')
+      DEFAULT_ENCODING = ENV.fetch('PLANTUML_ENCODING', 'legacy')
 
       attr_accessor :url, :txt_enable, :svg_enable, :png_enable, :encoding
 
@@ -79,26 +78,56 @@ module Asciidoctor
           txt_enabled? || png_enabled? || svg_enabled?
         end
 
-        def plantuml_content_format(code, format, attrs = {})
-          if %w[png svg].include?(format)
-            plantuml_img_content(code, format, attrs)
-          elsif format == 'txt' && txt_enabled?
-            plantuml_txt_content(code, format, attrs)
+        def plantuml_content_format(parent, code, format, attrs = {})
+          content = code.read
+
+          # honor subs attributes
+          # e.g. replace asciidoc variables
+          subs = attrs['subs']
+          content = parent.apply_subs(content, parent.resolve_subs(subs)) if subs
+
+          # add @start... and @end... if missing
+          content = "@startuml\n#{content}\n@enduml" unless content =~ /^@start.*@end[a-z]*$/m
+
+          # insert global plantuml config after first line
+          config_path = parent.attr('plantuml-include', '', true)
+
+          unless config_path.empty?
+            begin
+              source_file = parent.document.normalize_system_path(config_path, nil, nil, recover: false)
+              content = insert_config_to_content(parent, source_file, content, attrs)
+            rescue StandardError => e
+              return plantuml_invalid_file(source_file, e.message, attrs)
+            rescue SecurityError => e
+              return plantuml_insecure_file(source_file, e.message, attrs)
+            end
+          end
+
+          if %w[png svg txt].include?(format) && method("#{format}_enabled?").call
+            method("plantuml_#{format}_content").call(content, format, attrs)
           else
             plantuml_invalid_content(format, attrs)
           end
         end
 
-        def plantuml_content(code, attrs = {})
+        def plantuml_content(parent, code, attrs = {})
           format = attrs['format'] || DEFAULT_FORMAT
 
           return plantuml_disabled_content(code, attrs) unless enabled?
 
-          unless valid_uri?(server_url)
-            return plantuml_server_unavailable_content(server_url, attrs)
-          end
+          return plantuml_server_unavailable_content(server_url, attrs) unless valid_uri?(server_url)
 
-          plantuml_content_format(code, format, attrs)
+          plantuml_content_format(parent, code, format, attrs)
+        end
+
+        def plantuml_content_from_file(parent, target, attrs = {})
+          source_file = parent.document.normalize_system_path(target, nil, nil, recover: false)
+          content = ::File.open(source_file, mode: FILE_READ_MODE)
+          plantuml_content(parent, content, attrs)
+        rescue StandardError => e
+          plantuml_invalid_file(source_file, e.message, attrs)
+        rescue SecurityError => e
+          plantuml_insecure_file(source_file, e.message, attrs)
         end
 
         # Compression code used to generate PlantUML URLs. Taken directly from
@@ -117,15 +146,30 @@ module Asciidoctor
           join_paths(server_url, "#{format}/", result).to_s
         end
 
+        def create_plantuml_block(parent, content, attrs)
+          Asciidoctor::Block.new parent, :pass, {
+            content_model: :raw,
+            source: content,
+            subs: :default
+          }.merge(attrs)
+        end
+
         private
 
+        def insert_config_to_content(parent, config_path, content, attrs)
+          config = File.read(config_path, mode: FILE_READ_MODE)
+          subs = attrs['subs']
+          config = parent.apply_subs(config, parent.resolve_subs(subs)) if subs
+          return content.dup.insert(content.index("\n"), "\n#{config}") unless config.empty?
+        end
+
         def plantuml_txt_content(code, format, attrs = {})
           url = gen_url(code, format)
           URI(url).open do |f|
             plantuml_ascii_content(f.read, attrs)
           end
         rescue OpenURI::HTTPError, Errno::ECONNREFUSED, SocketError
-          plantuml_img_content(code, format, attrs)
+          plantuml_png_content(code, format, attrs)
         end
 
         def plantuml_ascii_content(code, attrs = {})
@@ -140,7 +184,7 @@ module Asciidoctor
           content + '</div>'
         end
 
-        def plantuml_img_content(code, format, attrs = {})
+        def plantuml_png_content(code, format, attrs = {})
           content = '<div class="imageblock">'
           content += '<div class="content">'
           content += '<img '
@@ -154,37 +198,53 @@ module Asciidoctor
           content + '</div>'
         end
 
-        def plantuml_invalid_content(format, attrs = {})
-          content = '<div class="listingblock">'
+        def plantuml_svg_content(code, format, attrs = {})
+          content = '<div class="imageblock">'
           content += '<div class="content">'
-          content += '<pre '
+          content += '<object type="image/svg+xml" '
+          content += "data=\"#{gen_url(code, format)}\" "
           content += "id=\"#{attrs['id']}\" " if attrs['id']
-          content += 'class="plantuml plantuml-error"> '
-          content += "PlantUML Error: Invalid format \"#{format}\""
-          content += '</pre>'
+          content += "width=\"#{attrs['width']}\" " if attrs['width']
+          content += "height=\"#{attrs['height']}\" " if attrs['height']
+          content += '>'
+          attrs['id'] = 'fallback_' + attrs['id'] if attrs['id']
+          content += plantuml_png_content(code, format, attrs)
+          content += '</object>'
           content += '</div>'
           content + '</div>'
         end
 
+        def plantuml_invalid_content(format, attrs = {})
+          error = "PlantUML Error: Invalid format \"#{format}\""
+          _plantuml_error_content(error, attrs)
+        end
+
         def plantuml_server_unavailable_content(url, attrs = {})
-          content = '<div class="listingblock">'
-          content += '<div class="content">'
-          content += '<pre '
-          content += "id=\"#{attrs['id']}\" " if attrs['id']
-          content += 'class="plantuml plantuml-error"> '
-          content += "Error: cannot connect to PlantUML server at \"#{url}\""
-          content += '</pre>'
-          content += '</div>'
-          content + '</div>'
+          error = "Error: cannot connect to PlantUML server at \"#{url}\""
+          _plantuml_error_content(error, attrs)
         end
 
         def plantuml_disabled_content(code, attrs = {})
+          _plantuml_error_content(code, attrs)
+        end
+
+        def plantuml_invalid_file(file, error, attrs = {})
+          error = "PlantUML Error: Could not parse \"#{file}\": #{error}"
+          _plantuml_error_content(error, attrs)
+        end
+
+        def plantuml_insecure_file(file, error, attrs = {})
+          error = "PlantUML Error: Could not read \"#{file}\": #{error}"
+          _plantuml_error_content(error, attrs)
+        end
+
+        def _plantuml_error_content(error, attrs = {})
           content = '<div class="listingblock">'
           content += '<div class="content">'
           content += '<pre '
           content += "id=\"#{attrs['id']}\" " if attrs['id']
           content += 'class="plantuml plantuml-error">\n'
-          content += code
+          content += error
           content += '</pre>'
           content += '</div>'
           content + '</div>'
@@ -244,11 +304,9 @@ module Asciidoctor
         end
 
         def expand_path(path, current, last, separator)
-          path = path[1..-1] if path.start_with?(separator) && current.zero?
+          path = path[1..] if path.start_with?(separator) && current.zero?
 
-          unless path.end_with?(separator) || current == last
-            path = [path, separator]
-          end
+          path = [path, separator] unless path.end_with?(separator) || current == last
 
           path
         end
@@ -259,31 +317,23 @@ module Asciidoctor
     class BlockProcessor < Asciidoctor::Extensions::BlockProcessor
       use_dsl
       named :plantuml
-      on_context :listing
+      on_context :listing, :literal
       content_model :simple
 
       def process(parent, target, attrs)
-        lines = target.lines
-
-        unless target.lines[0] =~ /@startuml/
-          lines = ['@startuml'] + target.lines
-        end
-
-        lines += ['@enduml'] unless target.lines[-1] =~ /@enduml/
-
-        content = Processor.plantuml_content(lines.join("\n"), attrs)
-
-        create_plantuml_block(parent, content, attrs)
+        content = Processor.plantuml_content(parent, target, attrs)
+        Processor.create_plantuml_block(parent, content, attrs)
       end
+    end
 
-      private
+    # PlantUML BlockMacroProcessor
+    class BlockMacroProcessor < Asciidoctor::Extensions::BlockMacroProcessor
+      use_dsl
+      named :plantuml
 
-      def create_plantuml_block(parent, content, attrs)
-        Asciidoctor::Block.new parent, :pass,  {
-          content_model: :raw,
-          source: content,
-          subs: :default
-        }.merge(attrs)
+      def process(parent, target, attrs)
+        content = Processor.plantuml_content_from_file(parent, target, attrs)
+        Processor.create_plantuml_block(parent, content, attrs)
       end
     end
   end
